#include	<stdio.h>

__global__ void kernel ( int * data, int n )
{
	size_t index = threadIdx.x + blockIdx.x * blockDim.x;

	if ( index < n )
		data [index]++;
}

int main ()
{
	int   * ptr;
	int	n = 160*1024*1024;		// # of elements
	int	blockSize = 1024;
	int	numBlocks = (n + blockSize - 1) / blockSize;
	int	device;

	cudaGetDevice ( &device );

	printf ( "Device %d\n", device );

					// allocate managed memory
	cudaMallocManaged ( &ptr, n * sizeof ( int ) );

	for ( int i = 0; i < n; i++ )
		ptr [i] = i;

	cudaMemPrefetchAsync ( ptr, n * sizeof(int), device );

	kernel<<<numBlocks, blockSize>>> ( ptr, n );

	cudaDeviceSynchronize ();

	for ( int i = 0; i < n; i++ )
		if ( ptr [i] != i + 1 )
		{
			printf ( "Error at %d - %d\n", i, ptr [i] );
			return 1;
		}

	cudaFree ( ptr );

	printf ( "Success\n" );

	return 0;
}
