#include	<stdio.h>
#include	<math.h>
 
// CUDA kernel to add elements of two arrays
__global__ void add ( int n, float * x, float * y )
{
	int index  = blockIdx.x * blockDim.x + threadIdx.x;
	int stride = blockDim.x * gridDim.x;
	
	for ( int i = index; i < n; i += stride )
		y [i] = x [i] + y [i];
}
int main ()
{
	int	device;

	cudaGetDevice ( &device );

	printf ( "Device %d\n", device );

	int	n = 1<<20;
	float * x, * y;
 
		// Allocate Unified Memory -- accessible from CPU or GPU
	cudaMallocManaged ( &x, n * sizeof(float) );
	cudaMallocManaged ( &y, n * sizeof(float) );
 
		// initialize x and y arrays on the host
	for (int i = 0; i < n; i++) 
	{
		x[i] = 1.0f;
		y[i] = 2.0f;
	}
 
		// Launch kernel on 1M elements on the GPU
	int blockSize = 256;
	int numBlocks = (n + blockSize - 1) / blockSize;

	add<<<numBlocks, blockSize>>> ( n, x, y );
 
		// Wait for GPU to finish before accessing on host
	cudaDeviceSynchronize();
 
		// Check for errors (all values should be 3.0f)
	float maxError = 0.0f;

	for (int i = 0; i < n; i++)
		maxError = fmax ( maxError, fabs(y[i]-3.0f) );

	printf ( "Max error: %f\n", maxError );
 
		// Free memory
	cudaFree(x);
	cudaFree(y);	

	return 0;
}
