#include	<stdio.h>
 
__global__ void kernel ( int n, int * x )
{
	int index  = blockIdx.x * blockDim.x + threadIdx.x;
	
	if ( index < n )
		x [index] += index;
}

int main ()
{
	int	device;

	cudaGetDevice ( &device );

	int	  n = 1<<25;
	int * a;
	int	  tileSize = 1024 * 128;		// 64K items per tile
	int	  numTiles = (n + tileSize - 1) / tileSize;
	int	  numRuns  = 10;

	printf ( "Device %d\nnumTiles %d\n", device, numTiles );

	
	cudaStream_t	s1, s2, s3;
	
	cudaStreamCreate ( &s1 );
	cudaStreamCreate ( &s2 );
	cudaStreamCreate ( &s3 );
	
			// Allocate Unified Memory -- accessible from CPU or GPU
	cudaMallocManaged ( &a, n * sizeof(int) );
 
			// initialize array a on the host
	for ( int i = 0; i < n; i++ ) 
		a [i] = 1;
 
			// do processing in tiles
	for ( int i = 0; i < numTiles; i++ ) 
	{ 
			// offload previous tile to the cpu if (i > 0) 
		if ( i > 0 )
			cudaMemPrefetchAsync ( a + tileSize * (i-1), tileSize * sizeof(int), cudaCpuDeviceId, s1 ); 

			// run multiple kernels on current tile 
		for ( int j = 0; j < numRuns; j++ ) 
			kernel<<<1024, 1024, 0, s2>>>( tileSize, a + tileSize * i ); 

			// prefetch next tile to the gpu 
		if ( i < numTiles ) 
			cudaMemPrefetchAsync ( a + tileSize * (i+1), tileSize * sizeof(int), device, s3 ); 

			// sync all streams 
		cudaDeviceSynchronize(); 
	} 
			// Check for errors (all values should be 3.0f)
		// Free memory
	cudaFree          ( a );
	cudaStreamDestroy ( s1 );
	cudaStreamDestroy ( s2 );
	cudaStreamDestroy ( s3 );
	
	printf ( "Sucess\n" );
	
	return 0;
}
