#include	<stdio.h>
__global__ void kernel ( int n, int * x )
{
	int index  = blockIdx.x * blockDim.x + threadIdx.x;
	
	if ( index < n )
		x [index] += index;
}

int main ()
{
	int	device;

	cudaGetDevice ( &device );

	int	  n = 1<<25;
	int * a;
	int	  tileSize = 1024 * 128;		// 64K items per tile
	int	  numTiles = (n + tileSize - 1) / tileSize;
	int	  numRuns  = 10;

	printf ( "Device %d\nnumTiles %d\n", device, numTiles );

	
	cudaStream_t	s1, s2, s3, st;
	cudaEvent_t		e1, e2, et;
	
	cudaStreamCreate ( &s1 );
	cudaStreamCreate ( &s2 );
	cudaStreamCreate ( &s3 );
	cudaEventCreate  ( &e1 );
	cudaEventCreate  ( &e2 );
	
			// Allocate Unified Memory -- accessible from CPU or GPU
	cudaMallocManaged ( &a, n * sizeof(int) );
 
			// initialize array a on the host
	for ( int i = 0; i < n; i++ ) 
		a [i] = 1;

			// prefetch first tile
	cudaMemPrefetchAsync ( a, tileSize * sizeof(int), 0, s2);
	cudaEventRecord      ( e1, s2 ); 

	for ( int i = 0; i < numTiles; i++ ) 
	{ 
			// make sure previous kernel and current tile copy both completed 
		cudaEventSynchronize ( e1 );  
		cudaEventSynchronize ( e2 );

			// run multiple kernels on current tile 
		for (int j = 0; j < numRuns; j++)
			kernel<<<1024, 1024, 0, s1>>> ( tileSize, a + tileSize * i ); 

		cudaEventRecord ( e1, s1 );

			// prefetch next tile to the gpu in a separate stream 
		if ( i < numTiles - 1 ) 
		{
				// make sure the stream is idle to force non-deferred HtoD prefetches first 
			cudaStreamSynchronize ( s2 );       
			cudaMemPrefetchAsync  ( a + tileSize * (i+1), tileSize * sizeof(int), 0, s2 ); 
			cudaEventRecord       ( e2, s2 );
		} 

			// offload current tile to the cpu after the kernel is completed using the deferred path 
		cudaMemPrefetchAsync ( a + tileSize * i, tileSize * sizeof(int), cudaCpuDeviceId, s1 );

			// rotate streams and swap events 
		st = s1; s1 = s2; s2 = st; 
		st = s2; s2 = s3; s3 = st; 
		et = e1; e1 = e2; e2 = et; 
	}
 
		// Free memory
	cudaFree          ( a  );
	cudaEventDestroy  ( e1 );
	cudaEventDestroy  ( e2 );
	cudaStreamDestroy ( s1 );
	cudaStreamDestroy ( s2 );
	cudaStreamDestroy ( s3 );
	
	printf ( "Sucess\n" );

	return 0;
}
